import logging
from collections.abc import Iterable
from functools import cache
from typing import Any

import pandas as pd
from clickhouse_driver import Client, dbapi
from nomenklatura.settings import STATEMENT_TABLE
from sqlalchemy import Select

from ftm_columnstore import settings

log = logging.getLogger(__name__)


def table_exists(e: Exception, table: str) -> bool:
    if f"Table default.{table} already exists" in str(e):
        return True
    if "projection with this name already exists" in str(e):
        return True
    return False


def get_compiled_query(q: Any) -> str:
    # FIXME: this is dangerous!
    if hasattr(q, "compile"):
        q = str(q.compile(compile_kwargs={"literal_binds": True}))
        q = q.replace("group_concat", "first_value")
    q = str(q)
    log.debug(q)
    return q


class Connection(dbapi.Connection):
    stream: bool = False

    def execute(self, q: Any, *args, **kwargs) -> dbapi.cursor.Cursor:
        cursor = self.cursor()
        q = get_compiled_query(q)
        cursor.execute(q, *args, **kwargs)
        return cursor

    def execution_options(self, *args, **kwargs) -> "Connection":
        self.stream = kwargs.get("stream_results", False)
        return self


class ClickhouseDialect:
    name = "postgres"


class ClickhouseEngine:
    def __init__(
        self,
        uri: str | None = settings.DATABASE_URI,
    ):
        self.dialect = ClickhouseDialect()  # FIXME
        self.name = "clickhouse"
        self.table = STATEMENT_TABLE
        self.table_fpx = f"{self.table}_fpx"
        self.table_xref = f"{self.table}_xref"
        self.view_stats = f"{self.table}_stats"
        self.view_fpx_freq = f"{self.table}_fpx_freq"
        self.tables = (
            self.table,
            self.table_fpx,
            self.table_xref,
            self.view_stats,
            self.view_fpx_freq,
        )
        self.uri = uri
        self.ensure(recreate=False, exists_ok=True)

    def __str__(self):
        return self.uri

    def __repr__(self):
        return f"<{self.__class__.__name__} ({self})>"

    def connect(self, use_numpy: bool | None = False) -> dbapi.Connection | Client:
        if use_numpy:
            uri = self.uri + "?use_numpy=True"
            return Client.from_url(uri)
        return Connection(self.uri)

    def ensure(self, recreate: bool | None = False, exists_ok: bool | None = False):
        with self.connect() as conn:
            if recreate:
                for stmt in self.drop_statements:
                    conn.execute(stmt)
            for stmt in self.create_statements:
                try:
                    conn.execute(stmt)
                except Exception as e:
                    if exists_ok and any(table_exists(e, t) for t in self.tables):
                        pass
                    else:
                        raise e
            # self.execute("GRANT ALL ON *.* TO CURRENT_USER WITH GRANT OPTION")

    def insert(self, df: pd.DataFrame, table: str | None = None) -> int:
        # https://clickhouse-driver.readthedocs.io/en/latest/features.html#numpy-pandas-support
        if df.empty:
            return 0
        table = table or self.table
        with self.connect(use_numpy=True) as conn:
            return conn.insert_dataframe("INSERT INTO %s VALUES" % table, df)

    def query_dataframe(self, query: Select) -> pd.DataFrame:
        query = get_compiled_query(query)
        with self.get_connection() as conn:
            return conn.query_dataframe(query)

    def sync(self):  # somehow not guaranteed by clickhouse
        with self.connect() as conn:
            conn.execute(f"OPTIMIZE TABLE {self.table} FINAL DEDUPLICATE")

    def optimize(self, full: bool | None = False):
        with self.connect() as conn:
            for table in (self.view_stats, self.view_fpx_freq):
                log.info(f"Optimizing `{table}` ...")
                conn.execute(f"OPTIMIZE TABLE {table} FINAL")
        if full:
            log.info(f"Optimizing `{self.table}` ...")
            self.sync()

    @property
    def create_statements(self) -> Iterable[str]:
        create_table = f"""
        CREATE TABLE {self.table}
        (
            `id`                      FixedString(40),
            `entity_id`               String,
            `canonical_id`            String,
            `prop`                    LowCardinality(String),
            `prop_type`               LowCardinality(String),
            `schema`                  LowCardinality(String),
            `value`                   String,
            `original_value`          Nullable(String),
            `dataset`                 LowCardinality(String),
            `lang`                    LowCardinality(String),
            `target`                  Boolean,
            `external`                Boolean,
            `first_seen`              Nullable(DateTime64),
            `last_seen`               DateTime64,
            INDEX cix (canonical_id) TYPE set(0) GRANULARITY 4,
            INDEX eix (entity_id) TYPE set(0) GRANULARITY 4,
            INDEX six (schema) TYPE set(0) GRANULARITY 1,
            INDEX dix (dataset) TYPE set(0) GRANULARITY 1,
            INDEX tix (prop_type) TYPE set(0) GRANULARITY 1,
            INDEX pix (prop) TYPE set(0) GRANULARITY 1
        ) ENGINE = ReplacingMergeTree(last_seen)
        PRIMARY KEY (canonical_id, entity_id, prop, value, id)
        ORDER BY (canonical_id, entity_id, prop, value, id)
        """

        create_table_fpx = f"""
        CREATE TABLE {self.table_fpx}
        (
            `algorithm`     Enum('fingerprint', 'metaphone1', 'metaphone2', 'soundex'),
            `value`         String,
            `dataset`       LowCardinality(String),
            `entity_id`     String,
            `schema`        LowCardinality(String),
            `prop`          LowCardinality(String),
            `prop_type`     LowCardinality(String),
            INDEX eix (entity_id) TYPE set(0) GRANULARITY 4,
            INDEX six (schema) TYPE set(0) GRANULARITY 1,
            INDEX tix (prop_type) TYPE set(0) GRANULARITY 1,
            INDEX pix (prop) TYPE set(0) GRANULARITY 1
        ) ENGINE = ReplacingMergeTree()
        PRIMARY KEY (algorithm,value,prop,schema,dataset)
        ORDER BY (algorithm,value,prop,schema,dataset,entity_id)
        """

        create_table_xref = f"""
        CREATE TABLE {self.table_xref}
        (
            `left_dataset`            String,
            `left_id`                 String,
            `left_schema`             LowCardinality(String),
            `left_country`            LowCardinality(String),
            `left_caption`            String,
            `right_dataset`           String,
            `right_id`                String,
            `right_schema`            LowCardinality(String),
            `right_country`           LowCardinality(String),
            `right_caption`           String,
            `judgement`               LowCardinality(String),
            `score`                   Decimal32(8),
            `ts`                      DateTime64,
            `user`                    String,
            INDEX lix (left_id) TYPE set(0) GRANULARITY 4,
            INDEX rix (right_id) TYPE set(0) GRANULARITY 4,
            INDEX ldix (left_dataset) TYPE set(0) GRANULARITY 1,
            INDEX rdix (right_dataset) TYPE set(0) GRANULARITY 1
        ) ENGINE = ReplacingMergeTree(ts)
        ORDER BY (left_id,right_id)
        """

        create_view_stats = f"""
        CREATE MATERIALIZED VIEW {self.view_stats} (
            dataset         LowCardinality(String),
            schema          LowCardinality(String),
            entities        AggregateFunction(count, UInt64),
            statements      AggregateFunction(count, UInt64)
        )
        ENGINE = AggregatingMergeTree()
        ORDER BY (dataset, schema)
        AS SELECT
            dataset,
            schema,
            countState(distinct canonical_id) AS entities,
            countState(*) AS statements
        FROM {self.table}
        GROUP BY dataset, schema
        """

        create_view_fpx_freq = f"""
        CREATE MATERIALIZED VIEW {self.view_fpx_freq} (
            value           String,
            freq            AggregateFunction(count, UInt32),
            len             UInt16
        )
        ENGINE = AggregatingMergeTree()
        ORDER BY (value)
        AS SELECT
            value,
            countState(value) AS freq,
            length(value) AS len
        FROM {self.table_fpx}
        WHERE algorithm = 'fingerprint'
        GROUP BY value
        """

        projections = (
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_dataset (
                SELECT * ORDER BY dataset,canonical_id,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_schema (
                SELECT * ORDER BY schema,canonical_id,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_dataset_schema (
                SELECT * ORDER BY dataset,schema,canonical_id,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_schema (
                SELECT * ORDER BY schema,canonical_id,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_values (
                SELECT * ORDER BY value,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_canonical_lookup (
                SELECT * ORDER BY entity_id,canonical_id)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_canonical_id (
                SELECT * ORDER BY canonical_id,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_entity_id (
                SELECT * ORDER BY entity_id,prop)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_prop_type (
                SELECT * ORDER BY prop_type,schema,dataset)""",
            f"""ALTER TABLE {self.table} ADD PROJECTION {self.table}_prop (
                SELECT * ORDER BY prop,schema,dataset)""",
            f"""ALTER TABLE {self.table_fpx} ADD PROJECTION {self.table_fpx}_value (
                SELECT * ORDER BY value,schema,dataset)""",
            f"""ALTER TABLE {self.table_xref} ADD PROJECTION {self.table_xref}_reverse (
                SELECT * ORDER BY
            right_dataset,right_schema,right_id,left_dataset,left_schema,left_id)""",
        )
        return (
            create_table,
            create_table_fpx,
            create_table_xref,
            create_view_stats,
            create_view_fpx_freq,
            *projections,
        )

    @property
    def drop_statements(self) -> str:
        return (
            f"DROP TABLE IF EXISTS {self.table}",
            f"DROP TABLE IF EXISTS {self.table_fpx}",
            f"DROP TABLE IF EXISTS {self.table_xref}",
            f"DROP VIEW IF EXISTS {self.view_stats}",
            f"DROP VIEW IF EXISTS {self.view_fpx_freq}",
        )


@cache
def get_engine(uri: str | None = None) -> ClickhouseEngine:
    uri = uri or settings.DATABASE_URI
    return ClickhouseEngine(uri)



from collections.abc import Generator, Iterable
from functools import cache, lru_cache
from typing import TypedDict, TypeVar

from followthemoney import model
from followthemoney.schema import Schema
from followthemoney.types import registry
from ftmq.types import CE
from nomenklatura.statement import Statement

from ftm_columnstore.phonetic import PhoneticAlgorithm, get_phonetics

FS = TypeVar("FS", bound="FingerprintStatement")

NAME_TYPE = str(registry.name)


class Fingerprint(TypedDict):
    algorithm: str
    value: str


class FingerprintStatement(Fingerprint):
    """A statement describing fingerprints and phonetic algorithms for an
    entity, useful for pre-matching blocks"""

    dataset: str
    entity_id: str
    schema: str
    prop: str
    prop_type: str

    @classmethod
    def from_row(cls, *values: str) -> "FingerprintStatement":
        return cls(zip(cls.__annotations__, values))


@cache
def get_schema(schema: str | Schema) -> Schema:
    return model.get(schema)


@lru_cache(1_000_000)
def fingerprint(value: str) -> list[Fingerprint]:
    fingerprints: list[Fingerprint] = []
    for algorithm in PhoneticAlgorithm:
        for v in get_phonetics(value, algorithm.value):
            fingerprints.append({"algorithm": algorithm.value, "value": v})
    return fingerprints


def should_fingerprint_stmt(stmt: Statement) -> bool:
    if not stmt.id or not stmt.schema or not stmt.value:
        return False
    schema = get_schema(stmt.schema)
    if schema.is_a("Mention") or schema.is_a("LegalEntity"):
        return stmt.prop_type == NAME_TYPE
    return False


def fingerprints_from_entity(entity: CE) -> Generator[FS, None, None]:
    yield from fingerprints_from_statements(entity.statements)


def fingerprints_from_statements(
    statements: Iterable[Statement],
) -> Generator[FS, None, None]:
    for stmt in statements:
        if should_fingerprint_stmt(stmt):
            for fp in fingerprint(stmt.value):
                if fp["value"]:
                    yield {
                        **{
                            "dataset": stmt.dataset,
                            "entity_id": stmt.entity_id,
                            "schema": stmt.schema,
                            "prop": stmt.prop,
                            "prop_type": stmt.prop_type,
                        },
                        **fp,
                    }


COLUMNS_FPX = tuple(FingerprintStatement.__annotations__.keys())



from enum import StrEnum
from functools import lru_cache
from typing import Literal

from fingerprints import generate as fp
from followthemoney.types import registry
from ftmq.types import CE
from libindic.soundex import Soundex
from metaphone import doublemetaphone
from normality import WS

SX = Soundex()


class PhoneticAlgorithm(StrEnum):
    fingerprint = "fingerprint"
    metaphone1 = "metaphone1"
    metaphone2 = "metaphone2"
    soundex = "soundex"


TPhoneticAlgorithm = Literal[
    PhoneticAlgorithm.fingerprint,
    PhoneticAlgorithm.metaphone1,
    PhoneticAlgorithm.metaphone2,
    PhoneticAlgorithm.soundex,
]

DEFAULT_PHONETIC_ALGORITHM = PhoneticAlgorithm.fingerprint


@lru_cache(10_000_000)
def tokenize(value: str) -> set[str]:
    tokens = set()
    tokens.add(value)
    tokens.update([t for t in value.split(WS) if len(t) > 5])  # FIXME
    return tokens


@lru_cache(10_000_000)
def get_fingerprint(value: str) -> str:
    return fp(value) or ""


@lru_cache(10_000_000)
def get_metaphone(value: str) -> tuple[str]:
    return tuple(x or "" for x in doublemetaphone(value))


@lru_cache(10_000_000)
def get_soundex(value: str) -> str:
    return SX.soundex(value) or ""


@lru_cache(10_000_000)
def get_phonetics(
    value: str, algorithm: TPhoneticAlgorithm | None = DEFAULT_PHONETIC_ALGORITHM
) -> tuple[str]:
    value = get_fingerprint(value)  # fingerprint always
    if not value:
        return ("",)
    tokens = tokenize(value)
    if algorithm == PhoneticAlgorithm.fingerprint:
        return tuple(t for t in tokens)
    if algorithm == PhoneticAlgorithm.metaphone1:
        return tuple(get_metaphone(t)[0] for t in tokens)
    if algorithm == PhoneticAlgorithm.metaphone2:
        return tuple(get_metaphone(t)[1] for t in tokens)
    if algorithm == PhoneticAlgorithm.soundex:
        return tuple(get_soundex(t) for t in tokens)


def get_entity_fpx(
    entity: CE,
    algorithm: TPhoneticAlgorithm | None = DEFAULT_PHONETIC_ALGORITHM,
) -> set[str]:
    values = set()
    for value in entity.get_type_values(registry.get("name")):
        values.update(get_phonetics(value, algorithm))
    for value in entity.get_type_values(registry.get("label")):
        values.update(get_phonetics(value, algorithm))
    return values



import logging
import warnings

import nomenklatura.settings
import numpy as np

from ftm_columnstore.engine import get_engine
from ftm_columnstore.store import get_store

# shut up
logging.getLogger("clickhouse_driver.columns.service").setLevel(logging.ERROR)
logging.getLogger("numpy.core.fromnumeric").setLevel(logging.ERROR)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

# FIXME sqlalchemy monkey patch not working
nomenklatura.settings.DB_URL = "sqlite:///:memory:"

__version__ = "0.3.1"

__all__ = ["get_engine", "get_store"]



from collections.abc import Generator
from functools import cache

import pandas as pd
from ftmq.model.dataset import C, Dataset
from ftmq.store import SQLStore
from nomenklatura import store as nk
from nomenklatura.dataset import DS
from nomenklatura.db import get_metadata
from nomenklatura.entity import CE
from nomenklatura.resolver import Resolver
from nomenklatura.statement import Statement, make_statement_table
from sqlalchemy import MetaData, select
from sqlalchemy.sql.selectable import Select

from ftm_columnstore.engine import get_engine
from ftm_columnstore.settings import BULK_WRITE_SIZE
from ftm_columnstore.statements import fingerprints_from_statements


class BaseClickhouseStore(nk.SQLStore):
    def __init__(
        self,
        dataset: DS,
        resolver: Resolver[CE],
        uri: str | None = None,
    ):
        super().__init__(dataset, resolver)
        self.metadata = MetaData()
        self.table = make_statement_table(self.metadata)
        self.engine = get_engine(uri)
        self.columns = [c.name for c in self.table.columns]

    def writer(self) -> nk.Writer[DS, CE]:
        return ClickhouseWriter(self)

    def view(self, scope: DS, external: bool = False) -> nk.View[DS, CE]:
        return nk.sql.SQLView(self, scope, external=external)

    def _iterate_stmts(
        self, q: Select, *args, **kwargs
    ) -> Generator[Statement, None, None]:
        for row in self._execute(q):
            data = dict(zip(self.columns, row))
            yield Statement.from_dict(data)


class ClickhouseStore(SQLStore, BaseClickhouseStore):
    pass


class ClickhouseWriter(nk.sql.SQLWriter[DS, CE]):
    BATCH_STATEMENTS = BULK_WRITE_SIZE

    def _upsert_batch(self) -> None:
        if self.batch:
            df = pd.DataFrame([s.to_dict() for s in self.batch])
            self.store.engine.insert(df, self.store.engine.table)
            df = pd.DataFrame(fingerprints_from_statements(self.batch))
            self.store.engine.insert(df, self.store.engine.table_fpx)
        self.batch = set()

    def pop(self, entity_id: str) -> list[Statement]:
        self.flush()
        table = self.store.table
        q = select(table).where(table.c.entity_id == entity_id)
        statements: list[Statement] = list(self.store._iterate_stmts(q))
        # q_delete = delete(table).where(table.c.entity_id == entity_id)
        # stmt = str(q_delete.compile(compile_kwargs={"literal_binds": True}))
        # self.store.engine.execute(stmt)
        return statements


@cache
def get_store(
    uri: str | None = None,
    catalog: C | None = None,
    dataset: Dataset | str | None = None,
    resolver: Resolver | str | None = None,
) -> ClickhouseStore:
    get_metadata.cache_clear()
    if isinstance(dataset, str):
        dataset = Dataset(name=dataset)
    return ClickhouseStore(catalog, dataset, uri=uri, resolver=resolver)



import os

VERSION = "0.3.1"


def get_env(name, default=None):
    value = os.environ.get(name)
    if value is not None:
        return str(value)
    if default is not None:
        return str(default)


DATABASE_URI = get_env("DATABASE_URI", "clickhouse://localhost/default")
LOG_LEVEL = get_env("LOG_LEVEL", "INFO")
BULK_WRITE_SIZE = int(get_env("BULK_WRITE_SIZE", 100_000))



import logging
from typing import Annotated, Optional

import typer
from ftmq.io import smart_read_proxies, smart_write_proxies
from rich import print

from ftm_columnstore import get_engine, get_store, settings

log = logging.getLogger(__name__)

cli = typer.Typer(no_args_is_help=True)


@cli.callback(invoke_without_command=True)
def cli_version(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False
):
    if version:
        print(settings.VERSION)
        raise typer.Exit()


@cli.command("init", help="Initialize database and table.")
def cli_init(
    recreate: Annotated[
        Optional[bool],
        typer.Option(
            ..., help="Recreate tables if existing (requires DROP TABLE privileges)"
        ),
    ] = False
):
    engine = get_engine()
    engine.ensure(recreate=recreate, exists_ok=True)


@cli.command("iterate", help="Read entities from the store.")
def cli_iterate(
    dataset: Annotated[
        Optional[str], typer.Option("-d", help="Dataset to read from")
    ] = None,
    out_uri: Annotated[
        str, typer.Option("-o", help="Entities uri (as interpreted by `ftmq`)")
    ] = "-",
):
    """
    Read entities from the store and write json to `-o` (default: stdout)
    """
    store = get_store(dataset=dataset)
    smart_write_proxies(out_uri, store.iterate(dataset), serialize=True)


@cli.command("write", help="Write entity fragments to the store.")
def cli_write(
    in_uri: Annotated[
        str, typer.Option("-i", help="Entities uri (as read by `ftmq`)")
    ] = "-",
    dataset: Annotated[
        Optional[str], typer.Option("-d", help="Dataset to write to")
    ] = None,
    optimize: Annotated[
        Optional[bool],
        typer.Option(
            ..., help="Optimize after import (don't do that when writing parallel)"
        ),
    ] = False,
):
    """
    Write json entities from `infile` to store.
    """
    store = get_store(dataset=dataset)
    with store.writer() as bulk:
        for ix, proxy in enumerate(smart_read_proxies(in_uri)):
            bulk.add_entity(proxy)
            if ix and ix % 10_000 == 0:
                print("Write entity `%d` ..." % ix)

    if optimize:
        engine = get_engine()
        engine.optimize()


@cli.command("optimize")
def cli_optimize(
    full: Annotated[
        Optional[bool],
        typer.Option(..., help="dedupe full"),
    ] = False,
):
    """
    Perform clickhouse table optimizations
    """
    engine = get_engine()
    engine.optimize(full)
